@@ -561,7 +561,7 @@ func run(appCtx *cli.Command, cfg config.Cfg, logger log.Logger) error {
 
 		// When transactions are disabled, set up dry-run statistics middleware
 		// to collect snapshot statistics without creating actual snapshots
-		dryRunLogCache, err := storagemgr.NewDryRunLogCache(30*time.Minute, 15000)
+		dryRunLogCache, err := storagemgr.NewDryRunLogCache(60*time.Minute, 15000)
 		if err != nil {
 			return fmt.Errorf("create dry-run log cache: %w", err)
 		}
@@ -771,6 +771,9 @@ func run(appCtx *cli.Command, cfg config.Cfg, logger log.Logger) error {
 	// database), but this is a close enough indication of startup latency.
 	logger.WithField("duration_ms", time.Since(beganRun).Milliseconds()).Info("Started Gitaly")
 
+	// Report leftover migration garbage directory statistics if there are any
+	migration.ReportLostFoundDirectoryExistence(logger, cfg)
+
 	if !cfg.DailyMaintenance.IsDisabled() {
 		shutdownWorkers, err := maintenance.StartWorkers(
 			ctx,
