@@ -6,6 +6,7 @@ import (
 	"errors"
 	"fmt"
 	"runtime"
+	"strings"
 	"sync"
 	"time"
 
@@ -15,6 +16,7 @@ import (
 	"gitlab.com/gitlab-org/gitaly/v16/internal/gitaly/storage/wal"
 	logging "gitlab.com/gitlab-org/gitaly/v16/internal/log"
 	"gitlab.com/gitlab-org/gitaly/v16/internal/safe"
+	"gitlab.com/gitlab-org/gitaly/v16/internal/structerr"
 	"gitlab.com/gitlab-org/gitaly/v16/proto/go/gitalypb"
 	"go.etcd.io/raft/v3"
 	"go.etcd.io/raft/v3/raftpb"
@@ -50,7 +52,7 @@ type RaftReplica interface {
 
 	// AddNode adds a new node to the Raft cluster.
 	// This operation can only be performed by the leader.
-	AddNode(ctx context.Context, address string) error
+	AddNode(ctx context.Context, address, destinationStorageName string) error
 
 	// RemoveNode removes a node from the Raft cluster.
 	// This operation can only be performed by the leader.
@@ -61,6 +63,32 @@ type RaftReplica interface {
 	// This is typically used to bring new nodes up to speed before promoting them to voters.
 	// This operation can only be performed by the leader.
 	AddLearner(ctx context.Context, address string) error
+
+	// GetCurrentState returns comprehensive current state information including term, index, and Raft state.
+	// This provides an efficient way to get multiple state values with consistent locking.
+	GetCurrentState() *ReplicaState
+}
+
+// StateString returns the normalized state string (removes "State" prefix and converts to lowercase).
+func StateString(state raft.StateType) string {
+	stateStr := state.String()
+	if strings.HasPrefix(stateStr, "State") {
+		return strings.ToLower(stateStr[5:])
+	}
+	return strings.ToLower(stateStr)
+}
+
+// ReplicaState holds comprehensive state information for a Raft replica.
+// This struct provides a consolidated view of the replica's current state,
+// including term, index, and Raft state, allowing for efficient querying
+// with consistent locking behavior.
+type ReplicaState struct {
+	// Term is the current Raft term from the live state machine
+	Term uint64
+	// Index is the current committed log index, equivalent to AppendedLSN
+	Index uint64
+	// State is the current Raft state (follower, candidate, leader, pre-candidate)
+	State raft.StateType
 }
 
 var (
@@ -151,6 +179,7 @@ type Replica struct {
 	cancel context.CancelFunc
 
 	memberID     uint64 // Member ID of the replica
+	relativePath string
 	partitionKey *gitalypb.RaftPartitionKey
 	node         raft.Node             // etcd/raft node representation
 	raftCfg      config.Raft           // etcd/raft configurations
@@ -204,6 +233,7 @@ func applyOptions(raftCfg config.Raft, opts []OptionFunc) (ReplicaOptions, error
 // RaftReplicaFactory defines a function type that creates a new Raft Replica instance.
 // This factory is used to create and initialize Replica objects for partitions.
 type RaftReplicaFactory func(
+	ctx context.Context,
 	memberID uint64,
 	storageName string,
 	partitionKey *gitalypb.RaftPartitionKey,
@@ -216,6 +246,7 @@ type RaftReplicaFactory func(
 // This function creates a Replica and registers it with the appropriate RaftEnabledStorage.
 func DefaultFactoryWithNode(raftCfg config.Raft, raftNode *Node, opts ...OptionFunc) RaftReplicaFactory {
 	return func(
+		ctx context.Context,
 		memberID uint64,
 		storageName string,
 		partitionKey *gitalypb.RaftPartitionKey,
@@ -233,7 +264,7 @@ func DefaultFactoryWithNode(raftCfg config.Raft, raftNode *Node, opts ...OptionF
 			return nil, fmt.Errorf("storage %q is not a RaftEnabledStorage", storageName)
 		}
 
-		replica, err := NewReplica(memberID, partitionKey, raftCfg, logStore, raftEnabledStorage, logger, metrics, opts...)
+		replica, err := NewReplica(ctx, memberID, partitionKey, raftCfg, logStore, raftEnabledStorage, logger, metrics, opts...)
 		if err != nil {
 			return nil, fmt.Errorf("create replica %q: %w", storageName, err)
 		}
@@ -252,6 +283,7 @@ func DefaultFactoryWithNode(raftCfg config.Raft, raftNode *Node, opts ...OptionF
 // start the Raft protocol. The Initialize method must be called separately to start
 // the Raft protocol operation.
 func NewReplica(
+	ctx context.Context,
 	memberID uint64,
 	partitionKey *gitalypb.RaftPartitionKey,
 	raftCfg config.Raft,
@@ -265,6 +297,11 @@ func NewReplica(
 		return nil, fmt.Errorf("raft is not enabled")
 	}
 
+	relativePath, err := storage.ExtractPartitioningHint(ctx)
+	if err != nil {
+		return nil, fmt.Errorf("extract partitioning hint: %w", err)
+	}
+
 	options, err := applyOptions(raftCfg, opts)
 	if err != nil {
 		return nil, fmt.Errorf("invalid raft replica option: %w", err)
@@ -274,6 +311,10 @@ func NewReplica(
 		return nil, fmt.Errorf("raft enabled storage is required")
 	}
 
+	if _, ok := raftEnabledStorage.GetTransport().(*GrpcTransport); !ok {
+		return nil, fmt.Errorf("transport is not a GrpcTransport")
+	}
+
 	logger = logger.WithFields(logging.Fields{
 		"component":         "raft",
 		"raft.partitionKey": partitionKey.GetValue(),
@@ -283,6 +324,7 @@ func NewReplica(
 
 	return &Replica{
 		memberID:           memberID,
+		relativePath:       relativePath,
 		partitionKey:       partitionKey,
 		raftCfg:            raftCfg,
 		options:            options,
@@ -553,6 +595,28 @@ func (replica *Replica) AppendedLSN() storage.LSN {
 	return replica.logStore.readCommittedLSN()
 }
 
+// GetCurrentState returns comprehensive current state information including term, index, and Raft state.
+// This provides a more efficient way to get multiple state values with consistent locking.
+func (replica *Replica) GetCurrentState() *ReplicaState {
+	replica.mutex.Lock()
+	defer replica.mutex.Unlock()
+
+	if replica.node == nil {
+		return &ReplicaState{
+			Term:  0,
+			Index: uint64(replica.logStore.readCommittedLSN()),
+			State: raft.StateFollower,
+		}
+	}
+
+	status := replica.node.Status()
+	return &ReplicaState{
+		Term:  status.Term,
+		Index: uint64(replica.logStore.readCommittedLSN()),
+		State: status.RaftState,
+	}
+}
+
 // LowWaterMark returns the earliest LSN that should be retained.
 // Log entries before this LSN can be safely removed.
 func (replica *Replica) LowWaterMark() storage.LSN {
@@ -863,9 +927,6 @@ func (replica *Replica) processConfChange(entry raftpb.Entry) error {
 	}
 
 	routingTable := replica.raftEnabledStorage.GetRoutingTable()
-	if routingTable == nil {
-		return fmt.Errorf("routing table not found")
-	}
 
 	// Apply the changes to the routing table
 	if err := routingTable.ApplyReplicaConfChange(replica.logStore.storageName, replica.partitionKey, replicaChanges); err != nil {
@@ -967,31 +1028,30 @@ func (replica *Replica) Step(ctx context.Context, msg raftpb.Message) error {
 }
 
 // AddNode implements RaftReplica.AddNode
-func (replica *Replica) AddNode(ctx context.Context, address string) error {
+func (replica *Replica) AddNode(ctx context.Context, address, destinationStorageName string) error {
 	memberID := uint64(replica.AppendedLSN() + 1)
-	return replica.proposeMembershipChange(ctx, string(addVoter), memberID, ConfChangeAddNode, &gitalypb.ReplicaID_Metadata{
+	return replica.proposeMembershipChange(ctx, string(addVoter), destinationStorageName, memberID, ConfChangeAddNode, &gitalypb.ReplicaID_Metadata{
 		Address: address,
 	})
 }
 
 // RemoveNode implements RaftReplica.RemoveNode
 func (replica *Replica) RemoveNode(ctx context.Context, memberID uint64) error {
-	return replica.proposeMembershipChange(ctx, string(removeNode), memberID, ConfChangeRemoveNode, nil)
+	return replica.proposeMembershipChange(ctx, string(removeNode), "", memberID, ConfChangeRemoveNode, nil)
 }
 
 // AddLearner implements RaftReplica.AddLearner
 func (replica *Replica) AddLearner(ctx context.Context, address string) error {
 	memberID := uint64(replica.AppendedLSN() + 1)
-	return replica.proposeMembershipChange(ctx, string(addLearner), memberID, ConfChangeAddLearnerNode, &gitalypb.ReplicaID_Metadata{
+	return replica.proposeMembershipChange(ctx, string(addLearner), "", memberID, ConfChangeAddLearnerNode, &gitalypb.ReplicaID_Metadata{
 		Address: address,
 	})
 }
 
-// proposeMembershipChange is a helper function that handles the common pattern for membership changes.
-// It checks leadership and proposes the configuration change.
 func (replica *Replica) proposeMembershipChange(
 	ctx context.Context,
-	changeType string,
+	changeType,
+	destinationStorageName string,
 	memberID uint64,
 	confChangeType ConfChangeType,
 	metadata *gitalypb.ReplicaID_Metadata,
@@ -1004,33 +1064,77 @@ func (replica *Replica) proposeMembershipChange(
 		return fmt.Errorf("replica is not the leader")
 	}
 
-	if confChangeType == ConfChangeRemoveNode {
-		routingTable := replica.raftEnabledStorage.GetRoutingTable()
-		if routingTable == nil {
-			return fmt.Errorf("routing table not found")
-		}
-		if err := checkMemberID(replica, memberID, routingTable); err != nil {
-			return fmt.Errorf("checking member ID: %w", err)
-		}
+	switch confChangeType {
+	case ConfChangeRemoveNode:
+		return replica.proposeRemoveNode(ctx, changeType, memberID)
+
+	case ConfChangeAddNode:
+		return replica.proposeAddNode(ctx, changeType, destinationStorageName, memberID, metadata)
+
+	case ConfChangeAddLearnerNode:
+		return replica.proposeAddLearner(ctx, changeType, memberID, metadata)
+
+	default:
+		return fmt.Errorf("config change type %v not supported", confChangeType)
+	}
+}
+
+func (replica *Replica) proposeRemoveNode(ctx context.Context, changeType string, memberID uint64) error {
+	// Validate member exists before removing
+	routingTable := replica.raftEnabledStorage.GetRoutingTable()
+	if err := checkMemberID(replica, memberID, routingTable); err != nil {
+		return err
 	}
 
-	w := replica.registry.Register()
+	return replica.proposeConfChange(ctx, changeType, memberID, ConfChangeRemoveNode, nil)
+}
 
+func (replica *Replica) proposeAddNode(ctx context.Context, changeType, destinationStorageName string, memberID uint64, metadata *gitalypb.ReplicaID_Metadata) (returnedErr error) {
+	// First, propose the configuration change
+	if err := replica.proposeConfChange(ctx, changeType, memberID, ConfChangeAddNode, metadata); err != nil {
+		return err
+	}
+
+	// After conf change commits, call JoinCluster on the target node
+	if err := replica.joinCluster(ctx, memberID, destinationStorageName, metadata); err != nil {
+		// If join fails, attempt to remove the node from the cluster
+		replica.logger.WithError(err).Warn("join cluster failed, attempting to remove node")
+
+		if removeErr := replica.proposeConfChange(ctx, string(removeNode), memberID, ConfChangeRemoveNode, nil); removeErr != nil {
+			replica.logger.WithError(removeErr).Error("failed to remove node after join failure")
+		}
+
+		return fmt.Errorf("join cluster failed: %w", err)
+	}
+
+	return nil
+}
+
+func (replica *Replica) proposeAddLearner(ctx context.Context, changeType string, memberID uint64, metadata *gitalypb.ReplicaID_Metadata) error {
+	return replica.proposeConfChange(ctx, changeType, memberID, ConfChangeAddLearnerNode, metadata)
+}
+
+func (replica *Replica) proposeConfChange(
+	ctx context.Context,
+	changeType string,
+	memberID uint64,
+	confChangeType ConfChangeType,
+	metadata *gitalypb.ReplicaID_Metadata,
+) (returnedErr error) {
+	waiter := replica.registry.Register()
 	defer func() {
+		replica.registry.Untrack(waiter.ID)
 		if returnedErr != nil {
-			replica.registry.Untrack(w.ID)
 			replica.metrics.IncMembershipError(changeType, "propose_failed")
 		}
 	}()
-
 	changes := NewReplicaConfChanges(
 		replica.node.Status().Term,
 		uint64(replica.AppendedLSN()),
 		replica.leadership.GetLeaderID(),
-		w.ID,
+		waiter.ID,
 		metadata,
 	)
-
 	changes.AddChange(memberID, confChangeType)
 
 	cc, err := changes.ToConfChangeV2()
@@ -1044,8 +1148,6 @@ func (replica *Replica) proposeMembershipChange(
 
 	replica.metrics.IncMembershipChange(changeType)
 
-	// Wait for the configuration change to be committed with a timeout
-	// If Raft silently drops the proposal (e.g., duplicate member ID), we need to timeout
 	timeout := time.Duration(replica.raftCfg.ProposalConfChangeTimeout) * time.Millisecond
 	timer := time.NewTimer(timeout)
 	defer timer.Stop()
@@ -1055,11 +1157,51 @@ func (replica *Replica) proposeMembershipChange(
 		return ctx.Err()
 	case <-timer.C:
 		return fmt.Errorf("configuration change timed out after %v - proposal may have been rejected by Raft", timeout)
-	case err := <-w.C:
+	case err := <-waiter.C:
 		return err
 	}
 }
 
+func (replica *Replica) joinCluster(ctx context.Context, targetMemberID uint64, destination string, targetMetadata *gitalypb.ReplicaID_Metadata) error {
+	if targetMetadata == nil || targetMetadata.GetAddress() == "" {
+		return fmt.Errorf("target node address is required")
+	}
+
+	routingTable := replica.raftEnabledStorage.GetRoutingTable()
+
+	routingEntry, err := routingTable.GetEntry(replica.partitionKey)
+	if err != nil {
+		return fmt.Errorf("getting routing table entry: %w", err)
+	}
+
+	transport := replica.raftEnabledStorage.GetTransport()
+	grpcTransport, ok := transport.(*GrpcTransport)
+	if !ok {
+		return fmt.Errorf("transport is not a GrpcTransport")
+	}
+
+	client, err := grpcTransport.getRaftClient(ctx, targetMetadata.GetAddress())
+	if err != nil {
+		return fmt.Errorf("get raft client for %s: %w", targetMetadata.GetAddress(), err)
+	}
+
+	_, err = client.JoinCluster(ctx, &gitalypb.JoinClusterRequest{
+		PartitionKey: replica.partitionKey,
+		LeaderId:     replica.leadership.GetLeaderID(),
+		MemberId:     targetMemberID,
+		Term:         replica.node.Status().Term,
+		Index:        replica.node.Status().Applied,
+		StorageName:  destination,
+		RelativePath: replica.relativePath,
+		Replicas:     routingEntry.Replicas,
+	})
+	if err != nil {
+		return structerr.NewInternal("join cluster RPC failed: %w", err)
+	}
+
+	return nil
+}
+
 func checkMemberID(replica *Replica, memberID uint64, routingTable RoutingTable) error {
 	_, err := routingTable.Translate(replica.partitionKey, memberID)
 	if err != nil {
